/*
The software in this package is distributed under the GNU General
Public License version 2 (with a special exception described below).

A copy of GNU General Public License (GPL) is included in this distribution,
in the file COPYING.GPL.

As a special exception, if other files instantiate templates or use macros
or inline functions from this file, or you compile this file and link it
with other works to produce a work based on this file, this file
does not by itself cause the resulting work to be covered
by the GNU General Public License.

However the source code for this file must still be made available
in accordance with section (3) of the GNU General Public License.

This exception does not invalidate any other reasons why a work based
on this file might be covered by the GNU General Public License.
*/
/** @file
 * @brief restricts html messages to an allowed group of tags.
 *
 * @copyright &copy; Copyright 2017 Intra2net AG
 *
 */

#include <string>
#include <sstream>
#include <iomanip>
#include <vector>

#include <boost/foreach.hpp>

#include <stringfunc.hxx>
#include <restricted_html.hpp>

using namespace std;

typedef pair<string,bool> TOKEN;

const vector<string> ALLOWED_PROTOCOLS = {"http://", "https://"};
const vector<string> ALLOWED_TAGS = {"h1", "h2", "h3", "h4", "h5", "h6",
                                     "a", "p", "br", "i", "ul", "li",
                                     "table", "tr", "th", "td"
                                    };
const string AHREF = "<a href=";
const string SAFE_URL_CHARS = "$-_.+!*'(),;/?:@=&abcdefghijklmnopqrstuvwxyz"
                              "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
const string REDIRECT_PREFIX = "/arnie?form=redirect&url=";
const string TARGET_BLANK = "target=_blank";


namespace I2n
{

    /**
     * @brief Compare protocol found in link against a whitelist.
     *
     * @param link url that will be checked.
     * @return true if link has an allowed protocol, false otherwise.
     */
    bool is_protocol_allowed(const string &link)
{
    BOOST_FOREACH(const string &protocol, ALLOWED_PROTOCOLS)
    {
        if (has_prefix(link, protocol))
            return true;
    }
    return false;
}

/**
 * @brief Verifies if a html "a" tag has a valid link and sanitize it if necessary.
 * Modify tag nd add redirector prefix if link has a valid protocol.
 * Example: <a href="http://somelink.com">
 * returns <a href="/arnie?form=redirect&url=http://somelink.com" target=_blank>
 *
 * @param tag html "a" tag.
 * @return true if link inside "a" tag is valid. False otherwise.
 */
bool link_sanitizer(string &tag)
{
    // tag = <a href="somelink">
    string link = tag.substr(AHREF.size());
    if (link.find_first_of("\"\'") == 0)
    {
        size_t pos = link.find_first_of("\"\'", 1);
        if (pos == string::npos)
            return false; // Quotation mark never closes.

        string end(link, pos+1);
        if (end.compare(" >") != 0 && end.compare(">") != 0)
            return false; //Probably extra attributes.

        link = link.substr(1, pos -1);
    }
    else
    {
        size_t space = link.find_first_of(" ");
        if (space != link.size()-2 && space != string::npos )
            return false; //Probably extra attributes.

        link = link.substr(0, space);
    }

    if (is_protocol_allowed(link))
        tag = AHREF + "\"" + REDIRECT_PREFIX + link + "\" " + TARGET_BLANK + ">";
    else if (link[0] != '/')
        return false;

    if (link.find_first_not_of(SAFE_URL_CHARS) != string::npos)
        return false;

    return true;
}

/**
 * @brief Check if tag is in a whitelist of allowed tags.
 * Does not accept a tag containing attributes.
 * Example:
 * <h1> or </p> returns true.
 * <br size="param"> returns false.
 *
 * @param tag html tag to be verified.
 * @return true if is an allowed tag, false otherwise.
 */
bool is_tag_allowed(string tag)
{
    replace_all(tag, "<", "");
    replace_all(tag, ">", "");
    replace_all(tag, "/", "");
    to_lower_mod(tag);

    BOOST_FOREACH(const string &a_tag, ALLOWED_TAGS)
    {
        if (tag.compare(a_tag) == 0)
            return true;
    }

    return false;
}

const string restrict_html(const string &html_code_orig, bool strip)
{
    string html_code(html_code_orig);
    remove_html_comments (html_code);
    vector<TOKEN > tokenized;
    tokenize_by_tag (tokenized, html_code);
    string result = "";

    vector<string> expected_tags;
    BOOST_FOREACH(const TOKEN &s, tokenized)
    {
        if (!s.second)
        {
            result = result + html_entities(s.first);
            continue;
        }
        if (is_tag_allowed(s.first) )
        {
            if (s.first.compare(0, 2, "</") != 0)
            {
                result = result + s.first;
                if (to_lower(s.first).compare("<br>") != 0)
                    expected_tags.push_back("</"+s.first.substr(1));
                continue;
            }
            else if (expected_tags.size() > 0 && expected_tags.back().compare(s.first) == 0)
            {
                result = result + s.first;
                expected_tags.pop_back();
                continue;
            }
        }

        string tag = s.first;
        if (to_lower(s.first).compare(0, AHREF.size(), AHREF) == 0 && link_sanitizer(tag))
        {
            result = result + tag;
            expected_tags.push_back("</a>");
            continue;
        }

        if (!strip)
            throw runtime_error("Invalid tag: " + s.first);

    }
    if (expected_tags.size() > 0) //One or more tags were not closed.
    {
        BOOST_REVERSE_FOREACH(const string &s, expected_tags)
        {
            result = result + s;
        }
    }

    return result.c_str();
}

/**
 * @brief Replace all "+" characters found in s to spaces (" ").
 *
 * @param s string that will be modified.
 */
static void unescape_space(string &s)
{
    string::size_type pos;
    while ((pos=s.find('+')) != string::npos)
        s[pos]=' ';
}

/**
 * @brief Converts a hexadecimal sequence to its respective character.
 *
 * @param s string of size 2. Example: "77"
 * @return respective character represented by the hex sequence.
 */
static char x2c(const string& s)
{
    char digit;
    digit=(s[0]>='A' ? ((s[0] & 0xdf)-'A')+10 : (s[0]-'0'));
    digit*=16;
    digit+=(s[1]>='A' ? ((s[1] & 0xdf)-'A')+10 : (s[1]-'0'));
    return digit;
}

/**
 * @brief Scan a string to find escaped hex chars in the format "%HH" and replace
 * for their respective character.
 * Example: "www%2E" becomes "www."
 *
 * @param s String that will be modified.
 */
static void unescape_hex(string& s)
{
    static char hex_escape='%';
    string::size_type escape_pos;
    string hex_seq;
    string rest=s;
    for (s=""; ((escape_pos=rest.find(hex_escape)) != string::npos);)
    {
        if (escape_pos+2<static_cast<unsigned int>(rest.length())
            && ::isalnum(rest[escape_pos+1]) && ::isalnum(rest[escape_pos+2]))
        {
            hex_seq=rest.substr(escape_pos+1,2);
            s=s+rest.substr(0,escape_pos)+x2c(hex_seq);
            rest=rest.erase(0,escape_pos+3);
        }
        else
        {
            s=s+rest.substr(0,escape_pos+1);
            rest=rest.erase(0,escape_pos+1);
        }
    }
    s+=rest;
}

/**
* @brief Decode url that contains percent-encoding. Replace space " " with "+".
* Example: "%77%77%77%2E" becomes "www."
*
* @param s url string.
* @return the decoded string.
*/
string decode_url(string s)
{
    unescape_space (s);
    unescape_hex (s);
    return (s);
}

/**
 * @brief Verify if the parameter character requires encoding, If it is non
 * alphanumeric or valid ascii signs.
 *
 * @param c character to be verified.
 * @return true if the character should be encoded.
 */
bool needs_encoding (const char &c)
{
    // some valid ascii signs
    if (c == '_' || c == '-')
        return false;

    // is digit?
    if (c > 47 && c < 58)
        return false;

    // is uppercase letter?
    if (c > 64 && c < 91)
        return false;

    // is lowercase letter?
    if (c > 96 && c < 123)
        return false;

    return true;
}

/**
* @brief Encode url with percent-encoding. Any non-alphanumeric character is
* converted to its hex value with the percent character (%) as prefix, except "_"
* and "-". Replace space " " with "+".
*
* @param s url string.
* @return the encoded url string.
*/
string encode_url(string s)
{
    // convert non-alphanumeric characters to hex, convert space to +
    ostringstream out;
    for (string::iterator pos2=s.begin(); pos2 != s.end(); pos2++)
    {
        if (*pos2 == ' ')
            out << '+';
        else if (needs_encoding (*pos2))
            out << '%' << std::uppercase << setw(2) << setfill('0') << \
                std::hex << (int)(unsigned char)*pos2;
        else
            out << (*pos2);
    }

    return out.str();
}
} // eo namespace I2n
